
def yaml_file_in_list(filename: str, file_paths: str) -> bool:
    """
    Checks whether a file name is present in the path

    Args:
        filename (str): File name
        file_paths (str): List of absolute file path

    Returns:
        Boolean - File is present in the path
    """
    return any(
        [filename in str(file_path) for file_path in file_paths]
    )



import logging
from typing import List, Union

import numpy as np

LOG = logging.getLogger(__name__)


class StringManager:
    @staticmethod
    def clean_str(messy):
        """
        Removes unwanted character in a str that we encounter alot
        """
        clean = messy

        # Strip of any chars that are beginning and end
        for ch in [' ', '\n']:
            clean = clean.strip(ch)

        # Remove colons but not when its between numbers (e.g time)
        if ':' in clean:
            work = clean.split(' ')
            result = []

            for w in work:
                s = w.replace(':', '')
                if s.isnumeric():
                    result.append(w)

                else:
                    result.append(s)

            clean = ' '.join(result)

        # Remove characters anywhere in string that is undesireable
        for ch in ['"', "'"]:
            clean = clean.replace(ch, '')

        clean = clean.strip(' ')
        return clean

    @staticmethod
    def get_encapsulated(str_line, encapsulator) -> List[str]:
        """
        Returns items found in the encapsulator, useful for finding units

        Args:
            str_line: String that has encapsulated info we want removed
            encapsulator: string of characters encapsulating info to be removed
        Returns:
            result: list of strings found inside anything between encapsulators

        e.g.
            line = 'density (kg/m^3), temperature (C)'
            ['kg/m^3', 'C'] = get_encapsulated(line, '()')
        """

        result = []

        if len(encapsulator) > 2:
            raise ValueError('encapsulator can only be 1 or 2 chars long!')

        elif len(encapsulator) == 2:
            lcap = encapsulator[0]
            rcap = encapsulator[1]

        else:
            lcap = rcap = encapsulator

        # Split on the lcap
        if lcap in str_line:
            for i, val in enumerate(str_line.split(lcap)):
                # The first one will always be before our encapsulated
                if i != 0:
                    if lcap != rcap:
                        result.append(val[0:val.index(rcap)])
                    else:
                        result.append(val)

        return result

    @classmethod
    def strip_encapsulated(cls, str_line: str, encapsulator: str) -> str:
        """
        Removes from a str anything that's encapusulated by characters and the
        encapsulating chars themselves

        Args:
            str_line: String that has encapsulated info we want removed
            encapsulator: string of characters encapsulating info to be removed
        Returns:
            final: String without anything between encapsulators
        """
        final = str_line
        result = cls.get_encapsulated(final, encapsulator)

        if len(encapsulator) == 2:
            lcap = encapsulator[0]
            rcap = encapsulator[1]

        else:
            lcap = rcap = encapsulator

        # Remove all the encapsulated words
        for v in result:
            final = final.replace(lcap + v + rcap, '')

        # Make sure we remove the last one
        return final

    @staticmethod
    def parse_none(value):
        """
        parses values looking for NANs, Nones, etc...

        Args:
            value: Value potentially containing a none or nan

        Returns:
            result: If string value is nan or none, then return None type otherwise
                    return original value
        """
        result = value

        # If its a nan or none or the string is empty
        if isinstance(value, str):
            if value.lower() in ['nan', 'none'] or not value:
                result = None
        elif isinstance(value, float):
            if np.isnan(value):
                result = None

        return result

    @classmethod
    def standardize_key(cls, messy):
        """
        Preps a key for use in dataframe columns or dictionary. Makes everything
        lowercase, removes units, replaces spaces with underscores.

        Args:
            messy: string to be cleaned
        Returns:
            clean: String minus all characters and patterns of no interest
        """
        key = messy

        # Remove units
        for c in ['()', '[]']:
            key = cls.strip_encapsulated(key, c)

        key = cls.clean_str(key)
        key = key.lower().replace(' ', '_')
        key = key.lower().replace('-', '_')

        # This removes csv byte order mark for files in utf-8
        # while were encoding with latin
        key = ''.join([c for c in key if c not in 'ï»¿'])

        return key

    @classmethod
    def infer_unit_from_key(cls, messy: str) -> Union[str, None]:
        """
        From a raw column name, infer the unit

        Args:
            messy: string to be cleaned
        Returns:
            unit: inferred unit if it exists
        """
        key = messy
        unit = None
        # Remove units
        for c in ['()', '[]']:
            result = cls.get_encapsulated(key, c)
            if result:
                unit = result[0].strip().lower()
                break
        return unit

    @classmethod
    def get_alpha_ratio(cls, str_line, encapsulator='""'):
        """
        Calculates the ratio of characters to numbers and
        potentially ignore things encapsulated

        Args:
            str_line: String to evaluate
            encapsulator: chars that encapsulate strings to be ignored

        Returns:
            ratio: float ratio of number of letter to number of numbers
        """

        line = str_line
        # Remove any quoted text
        if encapsulator:
            line = cls.strip_encapsulated(str_line, encapsulator='""')
        n_alpha = len([c for c in line if c.isalpha()])
        n_numeric = len([c for c in line if c.isnumeric()])

        if n_numeric == 0:
            ratio = 1
        else:
            ratio = n_alpha / n_numeric

        return ratio

    @classmethod
    def line_is_header(cls, str_line, header_sep=',', header_indicator='#',
                       previous_alpha_ratio=None, expected_columns=None):
        """
        Determine is line 1 a header line
        """
        # Definitive indication of a header line
        if header_indicator:
            return header_indicator == str_line[0]

        # No immediate answer so build confidence
        matches = []
        if previous_alpha_ratio:
            ratio = cls.get_alpha_ratio(str_line)
            matches.append(ratio >= previous_alpha_ratio)

        if header_sep:
            line = cls.strip_encapsulated(str_line, encapsulator='()')
            matches.append(len(line.split(header_sep)) == expected_columns)

        return matches.count(True) > matches.count(False)



URLS = [
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SSA.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_BSU_GPR.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_GM_SP.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SMP.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SD.001/",
    ("https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_GM_CSU_GPR.001/2020.02.06/"
     "SNEX20_GM_CSU_GPR_1GHz_v01.csv"),
    ("https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_UNM_GPR.001/2020.01.28/"
     "SNEX20_UNM_GPR.csv"),
    ("https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SD_TLI.001/2019.09.29/"
     "SNEX20_SD_TLI_clean.csv"),
]



from pathlib import Path

import pytest

from insitupy.variables import ExtendableVariables, \
    base_primary_variables_yaml


@pytest.fixture(scope="module")
def data_path():
    return Path(__file__).parent.joinpath(
        "data/snowex/pits/"
    ).expanduser().absolute()


@pytest.fixture(scope="session")
def base_primary_variables():
    return ExtendableVariables(entries=[base_primary_variables_yaml])


@pytest.fixture
def yaml_variable_file(tmp_path):
    """
    Create temporary yaml files for metadata or primary variables

    Yields:
        Function to create file
    """
    def _create_yaml_file(filename: str, variable_name: str = 'VAR_1'):
        file = tmp_path / filename
        file.write_text(
            f"{variable_name}:\n  description: Variable description"
        )
        return file

    yield _create_yaml_file



import pandas as pd
import pytest
import pytz

from insitupy.io.dates import DateTimeManager
from insitupy.io.yaml_codes import YamlCodes

in_timezone = 'US/Mountain'
out_timezone = 'UTC'


@pytest.fixture
def test_date():
    return pd.to_datetime('2025-04-01')


class TestDateTimeManager:
    # TODO: Cover GPR dates logic

    def test_parse_with_datetime(self):
        date = "2025-06-05T12:30:00"
        header = {
            YamlCodes.DATE_TIME: date,
        }

        result = DateTimeManager.parse(header)

        assert isinstance(result, pd.Timestamp)
        assert result == pd.Timestamp(date)

    def test_parse_with_separate_date_and_time(self, mocker):
        header = {
            YamlCodes.DATE: "2025-06-05",
            YamlCodes.TIME: "12:30:00",
        }
        mocked_handle_separate_datetime = mocker.patch(
            "insitupy.io.dates.DateTimeManager.handle_separate_datetime",
            return_value=pd.Timestamp("2025-06-05 12:30:00")
        )

        result = DateTimeManager.parse(header)

        mocked_handle_separate_datetime.assert_called_once_with(header)
        assert isinstance(result, pd.Timestamp)
        assert result == pd.Timestamp("2025-06-05 12:30:00")

    def test_parse_with_missing_data(self, mocker):
        header = {}
        mocked_handle_separate_datetime = mocker.patch(
            "insitupy.io.dates.DateTimeManager.handle_separate_datetime",
            return_value=None
        )

        result = DateTimeManager.parse(header)

        mocked_handle_separate_datetime.assert_called_once_with(header)
        assert result is None

    def test_parse_with_invalid_datetime(self):
        header = {
            YamlCodes.DATE_TIME: "invalid-datetime",
        }
        with pytest.raises(ValueError):
            DateTimeManager.parse(header)

    def test_separate_date_and_time(self):
        date = DateTimeManager.handle_separate_datetime(
            {
                'date': '2025-04-01',
                'time': '12:00',
                'header1': 'value1',
            }
        )
        assert type(date), pd.DatetimeIndex
        assert date.year == 2025
        assert date.month == 4
        assert date.day == 1
        assert date.hour == 12
        assert date.minute == 0

    def test_separate_date_no_time(self):
        date = DateTimeManager.handle_separate_datetime({'date': '2025-04-01'})
        assert type(date), pd.DatetimeIndex
        assert date.year == 2025
        assert date.month == 4
        assert date.day == 1
        assert date.hour == 00
        assert date.minute == 0

    def test_separate_date_time_nan(self):
        date = DateTimeManager.handle_separate_datetime(
            {'date': '2025-04-01', 'time': 'nan'}
        )
        assert type(date), pd.DatetimeIndex
        assert date.year == 2025
        assert date.month == 4
        assert date.day == 1
        assert date.hour == 00
        assert date.minute == 0

    def test_separate_date_mmddyy(self):
        date = DateTimeManager.handle_separate_datetime({'date': '040125'})
        assert type(date), pd.DatetimeIndex
        assert date.year == 2025
        assert date.month == 4
        assert date.day == 1
        assert date.hour == 00
        assert date.minute == 0

    def test_separate_date_mmddyy_time(self):
        date = DateTimeManager.handle_separate_datetime(
            {'date': '040125', 'time': '12:00'}
        )
        assert type(date), pd.DatetimeIndex
        assert date.year == 2025
        assert date.month == 4
        assert date.day == 1
        assert date.hour == 12
        assert date.minute == 0

    def test_separate_date_mmddyy_no_time(self):
        date = DateTimeManager.handle_separate_datetime(
            {'date': '040125', 'time': 'nan'}
        )
        assert type(date), pd.DatetimeIndex
        assert date.year == 2025
        assert date.month == 4
        assert date.day == 1
        assert date.hour == 00
        assert date.minute == 0

    def test_separate_date_no_date(self):
        with pytest.raises(ValueError):
            DateTimeManager.handle_separate_datetime({})

    # ========================= #
    # Method: adjust_timezone   #
    # ========================= #

    def test_adjust_timezone_date_with_no_zone(self, test_date):
        assert test_date.tz is None
        date = DateTimeManager.adjust_timezone(
            test_date, in_timezone, out_timezone
        )
        assert date.tz.zone == out_timezone
        # Should have shifted the hours. Not testing for exactness to make
        # the test safe against running on machines in different locations
        assert date.hour != test_date.hour

    def test_adjust_timezone_date_with_zone(self, test_date):
        test_date = test_date.tz_localize(pytz.timezone('US/Mountain'))
        date = DateTimeManager.adjust_timezone(
            test_date, in_timezone, out_timezone
        )
        assert date.tz.zone == out_timezone

    def test_adjust_timezone_no_in_zone(self, test_date):
        with pytest.raises(ValueError):
            DateTimeManager.adjust_timezone(
                date=test_date,
                out_timezone=out_timezone
            )



from insitupy.profiles.base import MeasurementData, ProfileData
from insitupy.variables.base_variables import MeasurementDescription
from insitupy.io.metadata import MetaDataParser


class TestProfileData:
    def test_inheritance(self):
        assert issubclass(ProfileData, MeasurementData) is True

    def test_init_with_default_arguments(self):
        profile_data = ProfileData()

        assert profile_data.variable is not None
        assert isinstance(profile_data.variable, MeasurementDescription)
        assert profile_data._meta_parser is not None
        assert isinstance(profile_data._meta_parser, MetaDataParser)

    def test_init_default_internal_properties(self):
        profile_data = ProfileData()

        assert [] == profile_data._measurements_to_keep
        assert [] == profile_data._non_measure_columns
        assert profile_data._has_layers is False

    def test_init_with_custom_arguments(self):
        variable = MeasurementDescription(
            code="DEPTH", description="Snow Depth"
        )
        meta_parser = MetaDataParser()
        profile_data = ProfileData(
            variable=variable, meta_parser=meta_parser
        )
        assert profile_data.variable == variable
        assert profile_data._meta_parser == meta_parser

    def test_default_depth_columns_initialization(self):
        profile_data = ProfileData()

        assert profile_data._depth_layer.code == "depth"
        assert profile_data._lower_depth_layer.code == "bottom_depth"

    def test_measurements_to_keep_initialization(self):
        variable = MeasurementDescription(
            code="DEPTH", description="Snow Depth"
        )
        meta_parser = MetaDataParser()
        profile_data = ProfileData(
            variable=variable, meta_parser=meta_parser
        )
        assert variable in profile_data._measurements_to_keep



from .snowex_metadata import SnowExMetaDataParser
from .snowex_campaign import SnowExProfileData
from .snowex_profile_data_collection import SnowExProfileDataCollection
from .variables import snowex_metadata_yaml, snowex_variables_yaml

__all__ = [
    "SnowExMetaDataParser",
    "SnowExProfileData",
    "SnowExProfileDataCollection",
    "snowex_metadata_yaml",
    "snowex_variables_yaml",
]



from unittest.mock import MagicMock, patch

import pytest

from insitupy.campaigns import ProfileDataCollection
from insitupy.io.metadata import MetaDataParser
from insitupy.profiles.base import ProfileData


@pytest.fixture
def mock_metadata_parser():
    mock_parser = MagicMock(spec=MetaDataParser)
    return mock_parser


@pytest.fixture
def mock_read_csv():
    with patch.object(ProfileDataCollection, '_read_csv') as mock_read_csv:
        mock_read_csv.return_value = ([], None)
        yield mock_read_csv


class TestProfileDataCollection:
    def test_profile_data_class(self):
        assert ProfileDataCollection.PROFILE_DATA_CLASS == ProfileData

    def test_from_csv_reads_csv_with_correct_params(self, mock_read_csv):
        filename = "test.csv"

        ProfileDataCollection.from_csv(filename=filename)

        mock_read_csv.assert_called_once()
        call_args = mock_read_csv.call_args[0]
        assert filename == call_args[0]
        assert isinstance(call_args[1], MetaDataParser)

    @patch(
        'insitupy.campaigns.ProfileDataCollection.'
        'PROFILE_DATA_CLASS.META_PARSER'
    )
    def test_from_csv_passes_params_to_meta_parser(
        self, mock_meta_parser, mock_read_csv
    ):
        filename = "test.csv"

        ProfileDataCollection.from_csv(filename=filename)

        mock_meta_parser.assert_called_once()
        assert mock_meta_parser.call_args[0][0] == 'US/Mountain'
        kwargs = mock_meta_parser.call_args.kwargs
        assert kwargs['primary_variable_file'] is None
        assert kwargs['metadata_variable_file'] is None
        assert kwargs['header_sep'] == ','
        assert kwargs['_id'] is None
        assert kwargs['campaign_name'] is None
        assert kwargs['allow_map_failures'] is False
        assert kwargs['allow_split_lines'] is True

    def test_from_csv_returns_collection_of_profiles(
        self, mock_metadata_parser, mock_read_csv
    ):
        metadata = MagicMock()
        profile1 = MagicMock()
        profile2 = MagicMock()
        mock_read_csv.return_value = ([profile1, profile2], metadata)

        result = ProfileDataCollection.from_csv(filename="test.csv")

        assert isinstance(result, ProfileDataCollection)
        assert result.profiles == [profile1, profile2]
        assert result.metadata == metadata

    def test_from_csv_ignores_profiles_with_variable_code_ignore(
        self, mock_metadata_parser, mock_read_csv
    ):
        metadata = MagicMock()
        profile1 = MagicMock(variable=MagicMock(code="DEPTH"))
        profile2 = MagicMock(variable=MagicMock(code="ignore"))
        mock_read_csv.return_value = ([profile1, profile2], metadata)

        result = ProfileDataCollection.from_csv(filename="test.csv")

        assert len(result.profiles) == 1
        assert result.profiles[0].variable.code == "DEPTH"



import numpy as np
import pytest

from insitupy.campaigns.snowex import SnowExProfileData
from insitupy.variables.base_variables import InputMappingError


class TestSnowexPitProfile:
    """
    Test the attributes of the profile
    """
    @pytest.mark.parametrize(
        "filename, variable", [
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                "PERMITTIVITY"
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
                "DENSITY"
            ),
        ]
    )
    def test_read_fails(
        self, filename, variable, data_path, base_primary_variables
    ):
        file_path = data_path.joinpath(filename)
        with pytest.raises(ValueError):
            SnowExProfileData(
                base_primary_variables.entries[variable],
            ).from_csv(
                file_path
            )

    @pytest.mark.parametrize(
        "filename, variable", [
            (
                "Density_file_unmapped.csv",
                "DENSITY"
            ),
        ]
    )
    def test_read_fails_mapping_error(
        self, filename, variable, data_path, base_primary_variables
    ):
        """
        Test that we fail to map a column in the column names when
        we expect them all to be mapped
        """
        file_path = data_path.joinpath(filename)
        with pytest.raises(InputMappingError):
            SnowExProfileData(
                base_primary_variables.entries[variable],
            ).from_csv(
                file_path
            )

    @pytest.mark.parametrize(
        "filename, variable, expected", [
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv",
                "SNOW_TEMPERATURE",
                0.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                "LWC_A",
                np.nan
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                "PERMITTIVITY_A",
                np.nan
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
                "DENSITY_A",
                397.8888888
            ),
        ]
    )
    def test_mean(
        self, filename, variable, expected, data_path, base_primary_variables
    ):
        file_path = data_path.joinpath(filename)
        obj = SnowExProfileData(base_primary_variables.entries[variable])
        obj.from_csv(file_path)
        result = obj.mean

        if np.isnan(expected):
            assert np.isnan(result)
        else:
            assert result == pytest.approx(expected)

    @pytest.mark.parametrize(
        "filename, variable, expected", [
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv",
                "SNOW_TEMPERATURE",
                95.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                "LWC_A",
                95.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                "PERMITTIVITY_B",
                95.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
                "DENSITY_A",
                95.0
            ),
        ]
    )
    def test_total_depth(
        self, filename, variable, expected, data_path, base_primary_variables
    ):
        file_path = data_path.joinpath(filename)
        obj = SnowExProfileData(base_primary_variables.entries[variable])
        obj.from_csv(file_path)
        result = obj.total_depth

        assert result == pytest.approx(expected)



from .variables import snowex_metadata_yaml, snowex_variables_yaml
from insitupy.io.metadata import MetaDataParser
from insitupy.variables import base_metadata_variables_yaml, \
    base_primary_variables_yaml


class SnowExMetaDataParser(MetaDataParser):
    """
    Base class for parsing SnowEx metadata header information and column names.
    """
    DEFAULT_METADATA_VARIABLE_FILES = [
        base_metadata_variables_yaml, snowex_metadata_yaml
    ]
    DEFAULT_PRIMARY_VARIABLE_FILES = [
        base_primary_variables_yaml, snowex_variables_yaml
    ]



import numpy as np
import pytest

from insitupy.campaigns import ProfileDataCollection
from insitupy.campaigns.snowex import SnowExProfileData, \
    SnowExProfileDataCollection

TEST_FILES = {
    "SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv":
        {
            'variables': ["SNOW_TEMPERATURE"],
            'means': [0.0]
        },
    "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv":
        {
            'variables':
                [
                    "DENSITY",
                    "PERMITTIVITY_A",
                    "PERMITTIVITY_B",
                    "LWC_A",
                    "LWC_B",
                ],
            'means': [395.03703703703707, np.nan, np.nan, np.nan, np.nan]
        },
    "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv":
        {
            'variables':
                [
                    "DENSITY_A",
                    "DENSITY_B",
                    "DENSITY_C"
                ],
            'means': [397.8888888, 394.5555556, 137.1111111]
        },
}


class TestSnowExProfileDataCollection:
    def test_inheritance(self):
        assert issubclass(SnowExProfileDataCollection, ProfileDataCollection) \
            is True

    def test_profile_data_class(self):
        assert SnowExProfileDataCollection.PROFILE_DATA_CLASS == \
            SnowExProfileData


@pytest.fixture
def data_collection(data_path):
    def _create_obj(filename):
        file_path = data_path.joinpath(filename)
        return SnowExProfileDataCollection.from_csv(
            file_path, allow_map_failure=True
        )
    yield _create_obj


@pytest.mark.parametrize('test_file', TEST_FILES)
class TestSnowExProfileDataCollectionFromCSV:
    def test_variables(
        self, test_file, data_collection, base_primary_variables
    ):
        obj = data_collection(test_file)
        result_variables = [p.variable for p in obj.profiles]
        expected_variables = [
            base_primary_variables.entries[entry]
            for entry in TEST_FILES[test_file]['variables']
        ]

        assert result_variables == expected_variables

    def test_mean(self, test_file, data_collection, base_primary_variables):
        obj = data_collection(test_file)
        result_means = [
            p.mean for p in obj.profiles
        ]
        np.testing.assert_almost_equal(
            np.array(result_means),
            np.array(TEST_FILES[test_file]['means']),
            decimal=2
        )



import logging

import geopandas as gpd
import numpy as np
import pandas as pd

from typing import Union

from insitupy.io.metadata import MetaDataParser
from insitupy.variables import MeasurementDescription

LOG = logging.getLogger(__name__)


class MeasurementData:
    """
    This would be one pit, SMP profile, etc
    Unique date, location, variable
    """
    META_PARSER = MetaDataParser

    def __init__(
        self,
        variable: MeasurementDescription = None,
        meta_parser: MetaDataParser = None,
    ):
        """
        Take df of layered data (SMP, pit)

        Args:
            variable: Description of variable. When None is given, then the
                definitions of the package YAML will be used.
            meta_parser: MetaDataParser object. This will hold our variable
                map and unit map. If no parser is given, then the default
                parser and definition of this package will be used.

        """
        if meta_parser is None:
            meta_parser = self.META_PARSER()
        self._meta_parser = meta_parser

        if variable is None:
            variable = MeasurementDescription()
        self.variable = variable

        # mapping of all column names to a measurement type
        self._column_mappings = {}
        self._sample_column = None

        self._df = None
        self._metadata = None
        # Columns that were identified in via MetaDataParser
        self._meta_columns_map = None

    def _set_column_mappings(self):
        # Get rid of columns we don't want and populate column mapping
        for column in self.columns:
            # Find the variable associated with each column
            # and store a map
            cn, cm = self._meta_parser.primary_variables.from_mapping(column)
            # join with existing mappings
            self._column_mappings = {**cm, **self._column_mappings}

    def _check_sample_columns(self):
        _sample_columns = [
            c for c in self.columns
            if self._column_mappings.get(c) == self.variable
        ]

        if len(_sample_columns) == 0:
            raise ValueError(
                f"No sample columns in {self.columns}. This is likely because"
                f" variable {self.variable} is not in columns"
                f" {self.columns}"
            )
        elif len(_sample_columns) > 1:
            LOG.warning(
                "Only one sample column allowed, keeping the first match"
            )
            _sample_columns = [_sample_columns[0]]

        _sample_column = _sample_columns[0]
        # Rename the variable column to the variable code
        # This covers the scenario where we did not auto-remap the
        # variable on original parsing of the column names because
        # it was a multi-sample variable
        # Columns related to the variable
        sample_column_type = self._column_mappings[_sample_column]
        if sample_column_type != self.variable:
            raise ValueError(
                f"{sample_column_type} and {self.variable} are not the same"
            )
        self._df.rename(
            columns={_sample_column: sample_column_type.code},
            inplace=True
        )
        self._sample_column = sample_column_type.code

    @property
    def metadata(self):
        return self._metadata

    @property
    def meta_columns_map(self):
        return self._meta_columns_map

    @metadata.setter
    def metadata(self, value):
        self._metadata = value

    @property
    def units_map(self):
        return self._meta_parser.units_map

    @property
    def latlon(self):
        # return location metadata
        return self._metadata.latitude, self._metadata.longitude

    @property
    def df(self):
        return self._df

    @df.setter
    def df(self, value):
        self._df = value

        if value.empty:
            LOG.warning("DF is empty")
        else:
            # This will populate the column mapping
            # and filter to the desired measurement column
            self._format_df()

    @property
    def columns(self) -> Union[np.ndarray, None]:
        """
        Gets the column names of the DataFrame if the DataFrame is not None.

        This property method retrieves the column names of the internal
        DataFrame stored in the `_df` attribute.

        Returns:
            numpy.ndarray or None: An array of column names
        """
        if self._df is None:
            return None
        else:
            return self._df.columns.values

    def _format_df(self):
        raise NotImplementedError("not implemented")

    @staticmethod
    def read_csv_dataframe(profile_filename, columns, header_position):
        """
        Read in a profile file. Managing the number of lines to skip and
        adjusting column names

        Args:
            profile_filename: Filename containing a manually measured
                             profile
            columns: list of columns to use in dataframe
            header_position: skiprows for pd.read_csv
        Returns:
            df: pd.dataframe of the csv data with desired column names
        """
        raise NotImplementedError("Not implemented")

    def from_csv(self, filename: str):
        """
        Parse all information of a given file, including the header and actual
        data.

        Args:
            filename: (str) Path of a file to read
        """
        # Parse the metadata and column info
        self._metadata, meta_columns, self._meta_columns_map, header_pos = \
            self._meta_parser.parse(filename=filename)

        # read in the actual data
        if meta_columns is None and not self._meta_columns_map:
            # Use an empty dataframe if the file is empty
            LOG.warning(f"File {filename} is empty of rows")
            self.df = pd.DataFrame()
        else:
            self.df = self.read_csv_dataframe(
                filename, meta_columns, header_pos
            )


class ProfileData(MeasurementData):
    """
    This would be one pit, SMP profile, etc
    Unique date, location, variable
    """
    NAN_DATA_VALUE = -9999

    def __init__(
        self,
        variable: MeasurementDescription = None,
        meta_parser: MetaDataParser = None,
    ):
        """
        See MeasurementData.__init__
        """
        # Init the measurement class
        super().__init__(variable=variable, meta_parser=meta_parser)

        # These are our default shared columns
        self._depth_columns = [
            self._meta_parser.primary_variables.entries["DEPTH"],
            self._meta_parser.primary_variables.entries["BOTTOM_DEPTH"]
        ]
        self._depth_layer = self._depth_columns[0]
        self._lower_depth_layer = self._depth_columns[1]
        # List of measurements to keep
        if variable is not None:
            self._measurements_to_keep = (
                self.shared_column_options() + [variable]
            )
        else:
            # If no variable is selected, then we will keep all measurements
            self._measurements_to_keep = []

        # Descriptive internal properties
        self._id = None
        self._dt = None
        self._has_layers = False
        self._non_measure_columns = []

    def shared_column_options(self):
        return self._depth_columns

    def _format_df(self):
        """
        Format the incoming df with the column headers and other info we want
        This will filter to a single measurement as well as the expected
        shared columns like "depth". This will also transform the dataframe
        into a GeoDataFrame with a geometry column from the parsed location in
        the metadata.
        """
        # Filter to the desired measurement columns
        self._set_column_mappings()
        if len(self._measurements_to_keep) > 0:
            columns_to_keep = [
                c for c in self.columns
                if self._column_mappings[c] in self._measurements_to_keep
            ]
            self._df = self._df.loc[:, columns_to_keep]

            # Verify the sample column exists and rename to the selected
            # variable
            self._check_sample_columns()

        self.describe()

        n_entries = len(self._df)
        self._df["datetime"] = [self._dt] * n_entries

        # Prepare location information
        lat, lon = self.latlon
        location = gpd.points_from_xy(
            [lon] * n_entries, [lat] * n_entries
        )

        self._df = gpd.GeoDataFrame(self._df, geometry=location).set_crs(
            "EPSG:4326"
        )
        self._df.replace(self.NAN_DATA_VALUE, np.NaN, inplace=True)

    def _add_thickness_to_df(self) -> None:
        """
        Calculates and adds the thickness column to the given dataframe if
        there is layer depth information.
        """
        self._df[
            self._meta_parser.primary_variables.entries[
                "LAYER_THICKNESS"
            ].code
        ] = (
            self._df[self._depth_layer.code] -
            self._df[self._lower_depth_layer.code]
        )

    def from_csv(self, filename: str):
        """
        See MeasurementData.from_csv
        """
        super().from_csv(filename)

        if len(self.columns) > 0 and self._depth_layer.code not in self.columns:
            raise ValueError(f"Expected {self._depth_layer} in columns")

    def describe(self) -> None:
        """
        Set internal properties based on csv file information and add some
        additional snow depth information if required information is present.
        """
        self._id = self.metadata.site_name
        self._dt = self.metadata.date_time

        # List of columns that are not the desired variable
        _non_measure_columns = [
            self._depth_layer.code,
            self._lower_depth_layer.code,
            "datetime",
            "geometry"
        ]
        # Remember them for later
        self._non_measure_columns = [
            c for c in _non_measure_columns if c in self.columns
        ]

        self._has_layers = self._lower_depth_layer.code in self.columns
        if self._has_layers:
            self._add_thickness_to_df()

    @property
    def sum(self):
        # get bulk value
        if not self._has_layers:
            # could we assume equidistant spacing and do a mean?
            raise RuntimeError("Cannot compute for no layers")

        # this should work for multi or not multi sample
        profile = self._df.loc[:, self._sample_column]
        self._df["mean"] = profile
        # TODO: sum up with depth change
        # TODO: could we use the weighted mean * the total depth?
        # TODO: units
        pass

    @property
    def mean(self):
        profile = self._df.loc[:, self._sample_column]
        if pd.isna(profile).all():
            return np.nan

        if self._has_layers:
            # height weighted mean for these layers
            thickness = self.df[
                self._meta_parser.primary_variables.entries[
                    "LAYER_THICKNESS"
                ].code
            ]
            # this works for a weighted mean, but is not assumed to be
            # the total thickness of the snowpack
            thickness_total = thickness.sum()
            weighted_mean = (
                profile * thickness / thickness_total
            ).sum()
            value = weighted_mean
        else:
            value = np.mean(profile)

        return value

    @property
    def total_depth(self):
        profile = self._df.loc[:, self._depth_layer.code].values
        return np.nanmax(profile)

    def get_profile(self, snow_datum="ground"):
        # TODO: snow datum is ground or snow
        # get profile of values
        profile_average = self._df.loc[:, self._sample_column]
        df = self._df.copy()
        df[self.variable.code] = profile_average
        columns_of_interest = [*self._non_measure_columns, self.variable.code]
        return df.loc[:, columns_of_interest]


def standardize_depth(depths, desired_format='snow_height', is_smp=False):
    """
    Data that is a function of depth comes in 2 formats. Sometimes 0 is
    the snow surface, sometimes 0 is the ground. This function standardizes it
    for each profile. Desired_format can be:

        snow_height: Zero at the bottom of the data.
        surface_datum: Zero at the top of the data and uses negative depths
                       (easier for plotting)

    Args:
        depths: Pandas series of depths in either format
        desired_format: string indicating which format the data is in
        is_smp: Boolean indicating which data this is, if smp then the data is
                surface_datum but with positive depths
   Returns:
        new:
    """
    max_depth = depths.max()
    min_depth = depths.min()

    new = depths.copy()

    # How is the depth ordered
    # max_depth_at_top = depths.iloc[0] > depths.iloc[-1]

    # Is the data in surface_datum already
    bottom_is_negative = depths.iloc[-1] < 0

    if desired_format == 'snow_height':

        if is_smp:
            LOG.info('Converting SMP depths to snow height format.')
            new = (depths - max_depth).abs()

        elif bottom_is_negative:
            LOG.info(
                'Converting depths in surface datum to snow height format.'
            )

            new = (depths + abs(min_depth))

    elif desired_format == 'surface_datum':
        if is_smp:
            LOG.info('Converting SMP depths to surface datum format.')
            new = depths.mul(-1)

        elif not bottom_is_negative:
            LOG.info(
                'Converting depths in snow height to surface datum format.'
            )
            new = depths - max_depth

    else:
        raise ValueError(
            f'{desired_format} is an invalid depth format! Options are:'
            f' {["snow_height", "surface_datum"]}'
        )

    return new



import pandas as pd
import pytest

from insitupy.campaigns.snowex import SnowExMetaDataParser
from insitupy.io.metadata import MetaDataParser
from tests.test_helpers import yaml_file_in_list


class TestSnowMetaDataParser:
    def test_inheritance(self):
        assert issubclass(SnowExMetaDataParser, MetaDataParser) is True

    def test_default_metadata_variables(self):
        assert yaml_file_in_list(
            'basemetadatavariables',
            SnowExMetaDataParser.DEFAULT_METADATA_VARIABLE_FILES
        )
        assert yaml_file_in_list(
            'snowexmetadatavariables',
            SnowExMetaDataParser.DEFAULT_METADATA_VARIABLE_FILES
        )

    def test_default_primary_variables(self):
        assert yaml_file_in_list(
            'baseprimaryvariables',
            SnowExMetaDataParser.DEFAULT_PRIMARY_VARIABLE_FILES
        )
        assert yaml_file_in_list(
            'snowexprimaryvariables',
            SnowExMetaDataParser.DEFAULT_PRIMARY_VARIABLE_FILES
        )


# List of test files with the values as the expected columns to be parsed
# successfully
TEST_FILES = {
    "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv":
        ['depth', 'bottom_depth', 'density_a', 'density_b', 'density_c'],
    "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv":
        ['depth', 'bottom_depth', 'density', 'permittivity_a',
         'permittivity_b', 'lwc_vol_a', 'lwc_vol_b'],
    "SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv":
        ['depth', 'snow_temperature']
}


@pytest.fixture(scope="class")
def meta_data_parser():
    return SnowExMetaDataParser(
        "US/Mountain",
        allow_map_failures=True
    )


@pytest.fixture(
    params=list(TEST_FILES.keys()),
    scope="class",
)
def metadata_info(request, meta_data_parser, data_path):
    metadata = meta_data_parser.parse(data_path.joinpath(request.param))
    return metadata, request.param


@pytest.fixture
def metadata(metadata_info):
    return metadata_info[0][0]


@pytest.fixture
def columns(metadata_info):
    return metadata_info[0][1], metadata_info[1]


@pytest.fixture
def header_pos(metadata_info):
    return metadata_info[0][3]


class TestSnowexPitMetadata:
    """
    Test that we can consistently read metadata across
    multiple pit measurements
    """

    def test_id(self, metadata):
        assert metadata.site_name == "COERAP_20200427_0845"

    def test_date_time(self, metadata):
        # converted from mountain time
        assert metadata.date_time == pd.to_datetime(
            "2020-04-27T14:45:00+0000"
        )

    def test_latitude(self, metadata):
        assert metadata.latitude == 38.92524

    def test_longitude(self, metadata):
        assert metadata.longitude == -106.97112

    def test_utm_epsg(self, metadata):
        assert metadata.utm_epsg == '26913'

    def test_site_name(self, metadata):
        assert metadata.campaign_name == "East River"

    def test_flags(self, metadata):
        assert metadata.flags is None

    def test_header_position(self, header_pos):
        assert header_pos == 10

    def test_expected_columns(self, columns):
        assert columns[0] == TEST_FILES[columns[1]]



from insitupy.io.metadata import MetaDataParser
from insitupy.variables import ExtendableVariables, \
    base_metadata_variables_yaml, base_primary_variables_yaml

from tests.test_helpers import yaml_file_in_list


class TestMetaDataParser:
    def test_default_metadata_variables(self):
        assert yaml_file_in_list(
            'basemetadatavariables',
            MetaDataParser.DEFAULT_METADATA_VARIABLE_FILES
        )

    def test_default_primary_variables(self):
        assert yaml_file_in_list(
            'baseprimaryvariables',
            MetaDataParser.DEFAULT_PRIMARY_VARIABLE_FILES
        )

    def test_default_initialization(self):
        """
        Test that the class initializes correctly with default arguments.
        """
        parser = MetaDataParser(timezone="UTC")
        assert parser._input_timezone == "UTC"
        assert parser._header_sep == MetaDataParser.DEFAULT_HEADER_SEPARATOR
        assert parser._campaign_name is None
        assert parser._id is None
        assert [str(base_primary_variables_yaml)] == \
               parser.primary_variables.source_files
        assert [str(base_metadata_variables_yaml)] == \
               parser.metadata_variables.source_files

    def test_custom_initialization(self, yaml_variable_file):
        """
        Test the class initializes with custom arguments.
        """
        primary_overwrite = yaml_variable_file('primary_overwrite.yaml')
        metadata_overwrite = yaml_variable_file('metadata_overwrite.yaml')
        parser = MetaDataParser(
            timezone="US/Mountain",
            primary_variable_file=str(primary_overwrite),
            metadata_variable_file=str(metadata_overwrite),
            header_sep="|",
            allow_split_lines=True,
            allow_map_failures=True,
            _id="1234",
            campaign_name="TestCampaign",
            units_map={"key": "value"}
        )
        assert parser._input_timezone == "US/Mountain"
        assert parser._header_sep == "|"
        assert parser._campaign_name == "TestCampaign"
        assert parser._id == "1234"
        assert parser._allow_split_header_lines is True
        assert str(primary_overwrite) in parser.primary_variables.source_files
        assert str(metadata_overwrite) in parser.metadata_variables.source_files
        assert parser._units_map == {"key": "value"}

    def test_extend_variables_only_default(self, yaml_variable_file):
        yaml_file = yaml_variable_file('variables.yaml')

        result = MetaDataParser.extend_variables([yaml_file])
        assert isinstance(result, ExtendableVariables)
        assert len(result) == 1
        assert not result.allow_map_failures

    def test_extend_variables_with_additions(self, yaml_variable_file):
        yaml_file = yaml_variable_file('variables.yaml')
        additions_file = yaml_variable_file('additions.yaml', 'META_1')

        result = MetaDataParser.extend_variables([yaml_file], additions_file)
        assert isinstance(result, ExtendableVariables)
        assert len(result) == 2

    def test_extend_variables_allow_map_failures(self, yaml_variable_file):
        yaml_file = yaml_variable_file('variables.yaml')

        result = MetaDataParser.extend_variables([yaml_file], allow_map_failures=True)
        assert isinstance(result, ExtendableVariables)
        assert len(result) == 1
        assert result.allow_map_failures



"""
Point data from select manual measurement campaigns
"""
import logging
from typing import List, Tuple

from insitupy.io.metadata import MetaDataParser, ProfileMetaData
from insitupy.profiles.base import ProfileData
from insitupy.variables import MeasurementDescription

LOG = logging.getLogger(__name__)


class ProfileDataCollection:
    """
    This could be a collection of profiles
    """
    PROFILE_DATA_CLASS = ProfileData

    def __init__(self, profiles: List[ProfileData], metadata: ProfileMetaData):
        self._profiles = profiles
        self._metadata = metadata

    @property
    def SWE(self):
        """
        Takes all layers for each unique date and location and returns point
        swe geo-dataframe. We can iterate though all ProfileData in this class
        and get SWE DF for each
        """
        # find all points with variable == density and calc SWE
        pass

    def get_mean(self, variable: MeasurementDescription):
        raise NotImplementedError()

    def get_sum(self, variable: MeasurementDescription):
        raise NotImplementedError()

    def get_profile(self, variable: MeasurementDescription):
        raise NotImplementedError()

    @property
    def metadata(self) -> ProfileMetaData:
        return self._metadata

    @property
    def profiles(self) -> List[ProfileData]:
        return self._profiles

    @classmethod
    def _read_csv(
        cls,
        filename,
        meta_parser: MetaDataParser,
        shared_column_options=None,
    ) -> Tuple[List[ProfileData], ProfileMetaData]:
        """
        Args:
            filename: path to csv
            meta_parser: metadata parser object
            shared_column_options: shared columns that will be used
                for data handling and storing. These come from primary
                variables but are not the primary data themselves

        Returns:
            a list of ProfileData objects

        """
        result = []

        # Parse the entire file before teasing out individual profiles
        all_profiles = cls.PROFILE_DATA_CLASS(meta_parser=meta_parser)
        all_profiles.from_csv(filename)

        # columns that will be included in data, but are not the primary
        # data themselves
        shared_column_options = shared_column_options or \
            all_profiles.shared_column_options()
        # Filter the shared columns
        shared_columns = [
            c for c, v in all_profiles.meta_columns_map.items()
            if v in shared_column_options
        ]
        # Variable columns are the remaining
        variable_columns = [
            c for c in all_profiles.meta_columns_map.keys()
            if c not in shared_columns
        ]

        # Create an object for each measurement
        for column in variable_columns:
            # Skip columns that are not mapped
            if all_profiles.meta_columns_map[column] is None:
                LOG.debug(f"Skipping column {column} because it is not mapped")
            else:
                profile = cls.PROFILE_DATA_CLASS(
                    meta_parser=meta_parser,
                    variable=all_profiles.meta_columns_map[column],
                )
                # IMPORTANT - Metadata needs to be set before assigning the
                # dataframe as information from the metadata is used to format_df
                # the information
                profile.metadata = all_profiles.metadata
                profile.df = all_profiles.df.loc[:, shared_columns + [column]].copy()
                # --------
                result.append(profile)
        if not result and all_profiles.df.empty:
            # Add one profile if this is empty so we can keep the metadata
            profile = cls.PROFILE_DATA_CLASS(
                meta_parser=meta_parser,
                variable=MeasurementDescription(),
            )
            profile.metadata = all_profiles.metadata
            result.append(profile)

        return result, all_profiles.metadata

    @classmethod
    def from_csv(
        cls,
        filename,
        timezone="US/Mountain",
        header_sep=PROFILE_DATA_CLASS.META_PARSER.DEFAULT_HEADER_SEPARATOR,
        site_id=None,
        campaign_name=None,
        allow_map_failure=False,
        metadata_variable_file=None,
        primary_variable_file=None
    ):
        """
        Find all profiles in a single csv file
        Args:
            filename: path to file
            timezone: expected timezone in file
            header_sep: header sep in the file
            site_id: Site id override for the metadata
            campaign_name: Campaign.name override for the metadata
            allow_map_failure: allow metadata and column unknowns
            metadata_variable_file:
                Optional addition to the recognized metadata variables defined
                in a YAML file
            primary_variable_file:
                Optional addition to the recognized primary variables defined
                in a YAML file
        Returns:
            This class with a collection of profiles and metadata
        """
        # TODO: timezone here (mapped from site?)
        meta_parser = cls.PROFILE_DATA_CLASS.META_PARSER(
            timezone,
            primary_variable_file=primary_variable_file,
            metadata_variable_file=metadata_variable_file,
            header_sep=header_sep,
            _id=site_id,
            campaign_name=campaign_name,
            allow_map_failures=allow_map_failure,
            allow_split_lines=True
        )

        profiles, metadata = cls._read_csv(filename, meta_parser)

        # ignore profiles with the name 'ignore'
        profiles = [
            p for p in profiles if
            # Keep the profile if it is None because we need the metadata
            (p.variable is None or p.variable.code != "ignore")
        ]

        return cls(profiles, metadata)



# tests/test_base_variables.py

import pytest

from insitupy.variables import ExtendableVariables, MeasurementDescription, \
    base_primary_variables_yaml
from insitupy.variables.base_variables import InputMappingError


@pytest.fixture
def sample_entries():
    return {
        "TEMP": MeasurementDescription(
            code="TEMP",
            description="Temperature measurement",
            map_from=["temperature", "temp"],
            auto_remap=True,
            match_on_code=True
        ),
        "DENSITY": MeasurementDescription(
            code="DENSITY",
            description="Density measurement",
            map_from=["density"],
            auto_remap=False,
            match_on_code=True
        )
    }


@pytest.fixture
def extendable_variables_fixture(sample_entries):
    return ExtendableVariables(entries=sample_entries)


class TestExtendableVariables:
    def test_variables_property(
        self, extendable_variables_fixture, sample_entries
    ):
        assert extendable_variables_fixture.variables == list(
            sample_entries.values()
        )

    def test_keys_property(
        self, extendable_variables_fixture, sample_entries
    ):
        assert extendable_variables_fixture.keys == list(sample_entries.keys())

    def test_iteration_over_variables(
        self, extendable_variables_fixture, sample_entries
    ):
        for i, variable in enumerate(extendable_variables_fixture):
            assert variable == list(sample_entries.values())[i]

    def test_length_of_entries(
        self, extendable_variables_fixture, sample_entries
    ):
        assert len(extendable_variables_fixture) == len(sample_entries)

    def test_from_mapping_success_on_code(
        self, extendable_variables_fixture
    ):
        result, mapping = extendable_variables_fixture.from_mapping("temp")
        assert result == "TEMP"
        assert result in mapping
        assert mapping[result].description == "Temperature measurement"

    def test_from_mapping_failure(self, extendable_variables_fixture):
        with pytest.raises(
                InputMappingError, match="Could not find mapping for humidity"
        ):
            extendable_variables_fixture.from_mapping("humidity")

    def test_from_mapping_allow_failures(self):
        ev = ExtendableVariables(entries={}, allow_map_failures=True)
        result, mapping = ev.from_mapping("unknown_variable")
        assert result == "unknown_variable"
        assert mapping[result] is None

    def test_source_files(self, yaml_variable_file):
        overwrites = yaml_variable_file('overwrites.yaml')
        ev = ExtendableVariables(
            entries=[base_primary_variables_yaml, overwrites],
            allow_map_failures=True
        )
        assert len(ev.source_files) == 2
        assert str(overwrites) in ev.source_files
        assert str(base_primary_variables_yaml) in ev.source_files

        # Make sure we properly append the file entries to the list of files
        # and don't maintain old entries between new object instances
        overwrites = yaml_variable_file('overwrites_2.yaml')
        ev = ExtendableVariables(
            entries=[overwrites],
            allow_map_failures=True
        )
        assert len(ev.source_files) == 1
        assert ev.source_files == [str(overwrites)]

    def test_to_dict(self, extendable_variables_fixture):
        result = extendable_variables_fixture.to_dict()
        assert isinstance(result, dict)
        assert "TEMP" in result
        assert result["TEMP"]["description"] == "Temperature measurement"

    def test_to_yaml(self, extendable_variables_fixture, tmp_path):
        yaml_file = tmp_path / "output.yaml"
        extendable_variables_fixture.to_yaml(yaml_file)
        assert yaml_file.exists()
        with open(yaml_file, "r") as f:
            content = f.read()
            assert "TEMP" in content
            assert "description: Temperature measurement" in content


class TestMeasurementDescription:
    @pytest.mark.parametrize(
        "code, description, map_from, auto_remap, match_on_code, cast_type",
        [
            ("AIR_TEMP", "Temperature", ["temp", "temperature"],
             True, True, "int"),
            ("", None, [], False, True, None),
            ("PRESS", "Pressure", ["pressure"], False, False, "float"),
        ],
    )
    def test_initialization(
        self, code, description, map_from, auto_remap, match_on_code, cast_type
    ):
        measurement = MeasurementDescription(
            code=code,
            description=description,
            map_from=map_from,
            auto_remap=auto_remap,
            match_on_code=match_on_code,
            cast_type=cast_type,
        )
        assert measurement.code == code
        assert measurement.description == description
        assert measurement.map_from == map_from
        assert measurement.auto_remap == auto_remap
        assert measurement.match_on_code == match_on_code
        assert measurement.cast_type == cast_type

    def test_default_values(self):
        measurement = MeasurementDescription()
        assert measurement.code == "-1"
        assert measurement.description is None
        assert measurement.map_from == []
        assert not measurement.auto_remap
        assert measurement.match_on_code
        assert measurement.cast_type is None

    def test_invalid_code_type(self):
        with pytest.raises(TypeError):
            MeasurementDescription(code=123)

    def test_invalid_map_from_type(self):
        with pytest.raises(TypeError):
            MeasurementDescription(map_from="not a list")



from pathlib import Path

from .base_variables import ExtendableVariables, MeasurementDescription

here = Path(__file__).parent.resolve()
base_primary_variables_yaml = here / "./baseprimaryvariables.yaml"
base_metadata_variables_yaml = here / "./basemetadatavariables.yaml"

__all__ = [
    "ExtendableVariables",
    "MeasurementDescription",
    "base_metadata_variables_yaml",
    "base_primary_variables_yaml"
]



import logging
import os.path
from pathlib import Path
from typing import Dict, List, Tuple, Union

import attrs
import pydash
import yaml
from attrs import field, validators

LOG = logging.getLogger(__name__)


class InputMappingError(RuntimeError):
    """
    Indicate we failed to map a column or header name
    """
    pass


# Similar to MeasurementDescription from metloom
@attrs.frozen
class MeasurementDescription:
    """
    data class for describing a measurement

    Args:
        code: code used within the applicable API
        description: description of the sensor
        map_from: List map to this variable from a list of options
        auto_remap: Auto remap to the column to the code
        match_on_code: Match on the code too
        cast_type: make this float, int, etc
    """
    # Code used within the applicable API
    code: str = field(default='-1', validator=validators.instance_of(str))
    # Description of the measurement
    description: str = None
    # Map to this variable from a list of options
    map_from: List = field(
        default=[],
        validator=attrs.validators.optional(validators.instance_of(List))
    )
    # Auto remap the column to the code
    auto_remap: bool = False
    # Match on the code too
    match_on_code: bool = True
    # Optional value type casting
    cast_type: str = None


def variable_from_input(files: list[Union[str, Path]], self_):
    """
    Parses list of YAML files that have primary or metadata variable
    definitions.

    Args:
        files (list[Union[str, Path]] | dict): List of files to parse.
        self_: The instance of the initialized ExtendableVariables class.

    Returns:
        dict: A dictionary with parsed MeasurementDescription objects.

    Raises:
        TypeError: If the input `x` is neither a list of files nor a valid
        dictionary.
    """
    # TODO: better logic here
    if isinstance(files, list) and all(os.path.isfile(f) for f in files):
        data_final = {}
        # If we have a list of files, we need to import them
        for f in files:
            self_.source_files = self_.source_files + [str(f)]
            with open(f) as fp:
                data = yaml.safe_load(fp)
            # Merge, overwriting options with second
            pydash.merge(data_final, data)
        return {k: MeasurementDescription(**v) for k, v in data_final.items()}
    # Check that we have a dict
    if not isinstance(files, dict):
        raise TypeError(
            f"Expected to formulate dict, got {type(files)} with value {files}"
        )
    return files


@attrs.define
class ExtendableVariables:
    """
    Make a class with the helpful iterator for storing variable options
    """
    source_files: list[str] = []
    entries: dict[str, MeasurementDescription] = attrs.field(
        factory=dict,
        converter=attrs.Converter(variable_from_input, takes_self=True)
    )
    allow_map_failures: bool = False

    @property
    def variables(self):
        return list(self.entries.values())

    @property
    def keys(self):
        return list(self.entries.keys())

    def __iter__(self):
        for item in self.variables:
            yield item

    def __len__(self):
        return len(self.entries)

    def from_mapping(
        self, input_name
    ) -> Tuple[str, Dict[str, MeasurementDescription]]:
        """
        Get the measurement description from an input name.
        This will use the MeasurementDescription.map_from list to
        check, in order, which variable we should use

        Args:
            input_name: string input name

        Returns:
            column name
            column mapping (map of name to MeasurementDescription)
        """
        lower_name = input_name.lower()
        result = None

        # Map column name to variable type
        column_mapping = {}
        for entry in self.variables:
            # Check if we match directly on the code
            code_match = entry.match_on_code and lower_name == entry.code
            # Check if we match from the list of possible matches
            map_match = entry.map_from and lower_name in entry.map_from
            if code_match or map_match:
                # Remap to code
                if entry.auto_remap:
                    result = entry.code
                else:
                    result = input_name
                # store a map of the column name to the variable description
                column_mapping[result] = entry
                break

        if result is None:
            if self.allow_map_failures:
                # We failed to find a mapping, but want to continue
                LOG.warning(f"Could not find mapping for {input_name}")
                result = input_name
                column_mapping[result] = None
            else:
                raise InputMappingError(
                    f"Could not find mapping for {input_name}"
                )
        LOG.debug(
            f"Mapping {result} to {result} (type {column_mapping[result]})"
        )
        return result, column_mapping

    def to_dict(self):
        return {k: attrs.asdict(v) for k, v in self.entries.items()}

    def to_yaml(self, output_file):
        with open(output_file, 'w') as fp:
            data = self.to_dict()
            yaml.dump(data, fp)



from insitupy.campaigns import ProfileDataCollection
from .snowex_campaign import SnowExProfileData


class SnowExProfileDataCollection(ProfileDataCollection):
    PROFILE_DATA_CLASS = SnowExProfileData



import numpy as np
import pytest

from insitupy.io.strings import StringManager


@pytest.mark.parametrize(
    "in_str, expected", [
        ('SMP instrument #', 'smp_instrument_#'),
        ('Dielectric Constant A', 'dielectric_constant_a'),
        ('Specific surface area (m^2/kg)', 'specific_surface_area'),
        # Ensure we remove a csv byte order mark in latin encoding
        ("ï»¿Camera", "camera"),
        (' Temperature \n', 'temperature')
    ]
)
def test_standardize_key(in_str, expected):
    """
    Test whether we can clean out the column header from a csv and standardize them
    """
    assert StringManager.standardize_key(in_str) == expected


@pytest.mark.parametrize('args, kwargs', [
    # Test multiple values being returned
    (
        ['Density [kg/m^3], Date [yyyymmdd]', '[]', ['kg/m^3', 'yyyymmdd']],
        {'errors': False}
    ),
    # Test single value being return with parenthese
    (['Time (seconds)', '()', ['seconds']], {'errors': False}),
    # Test Single encapsulator used as both
    (['Name "Surveyor"', '"', ['Surveyor']], {'errors': False}),
    # Test nothing returned
    (['Name', '()', []], {'errors': False}),
    # Test our value error for incorrect encaps
    (['Name', '()()', []], {'errors': True}), ])
def test_get_encapsulated(args, kwargs):
    """
    Test where we can remove chars in a string
    """
    s, encaps, expected = args
    # Errors out to test exception
    errors = kwargs['errors']

    if not errors:
        results = StringManager.get_encapsulated(s, encaps)
        for r, e in zip(results, expected):
            assert r == e
    else:
        with pytest.raises(ValueError):
            results = StringManager.get_encapsulated(s, encaps)


@pytest.mark.parametrize(
    's, encaps, expected', [
        ('Density [kg/m^3], Date [yyyymmdd]', '[]', 'Density , Date '),
        ('Time (seconds)', '()', 'Time '),
        ('Name "Surveyor"', '"', 'Name '),
        # test for mm and comments exchange
        ('grain_size (mm), comments', '()', 'grain_size , comments')
    ])
def test_strip_encapsulated(s, encaps, expected):
    """
    Test where we can remove chars in a string
    """
    r = StringManager.strip_encapsulated(s, encaps)
    assert r == expected


def test_parse_none():
    """
    Test we can convert nones and nans to None and still pass through everything
    else.
    """
    # Assert these are converted to None
    for v in ['NaN', '', 'NONE', np.nan]:
        assert StringManager.parse_none(v) is None

    # Assert these are unaffected by function
    for v in [10.5, 'Comment']:
        assert StringManager.parse_none(v) == v


@pytest.mark.parametrize("str_line, encapsulator, expected", [
    # Test simple 50/50
    ("1A", None, 1),
    # Use the ignore encapsulator
    ('1"A"', '""', 0),
    # Check for the div by zero
    ('A', None, 1),

])
def test_get_alpha_ratio(str_line, encapsulator, expected):
    result = StringManager.get_alpha_ratio(str_line, encapsulator=encapsulator)
    assert result == expected


@pytest.mark.parametrize(
    "line, header_sep, header_indicator, previous_alpha_ratio,"
    "expected_columns, expected", [
        # Test using a header indicator is the end all.
        ("# flags, ", None, '#', None, None, True),
        ("flags, ", None, '#', None, None, False),

        # Simple looking for 2 columns
        ("# flags, ", ',', None, None, 2, True),
        # Test with a real entry from stratigraphy which has lots of chars
        ("35.0,33.0,< 1 mm,DF,F,D,NaN", ',', '#', 7, 0.5, False),
        # Test a complicated string with encapsulation right after a normal header
        ('107.0,85.0,< 1 mm,FC,4F,D,"Variable."', ',', '#', 7, 1, False),
    ])
def test_line_is_header(
    line, header_sep, header_indicator, previous_alpha_ratio,
    expected_columns, expected
):
    result = StringManager.line_is_header(
        line, header_sep, header_indicator, previous_alpha_ratio,
        expected_columns
    )
    assert result == expected



from datetime import timedelta

import pandas as pd
import pytz

from .strings import StringManager
from .yaml_codes import YamlCodes


class RowKeys:
    UTC_DOY = 'utcdoy'
    UTC_YEAR = 'utcyear'
    UTC_TOD = 'utctod'


class DateTimeManager:
    @staticmethod
    def parse(rows: dict) -> pd.Timestamp:
        """
        Parses the header dictionary to extract a datetime value.

        Args:
            rows (dict): Dictionary containing date and time information.

        Returns:
            pd.Timestamp: A pandas Timestamp object representing the parsed
            datetime value.
        """
        datetime = None
        # Look for a 'datetime' header entry
        if rows.get(YamlCodes.DATE_TIME) is not None:
            str_date = str(
                rows[YamlCodes.DATE_TIME].replace('T', '-')
            )
            datetime = pd.to_datetime(str_date)

        # If we didn't find date/time combined.
        if datetime is None:
            datetime = DateTimeManager.handle_separate_datetime(rows)

        return datetime

    @staticmethod
    def handle_separate_datetime(rows):
        """
        Handle a separate date and time entry. This assumes the keys
        'date' and 'time' in two separate lines in the header.

        Args:
            rows: parsed header rows
        Returns:
            parsed datetime
        """
        keys = [k.lower() for k in rows.keys()]

        # Handle data dates and times
        if YamlCodes.DATE in keys and YamlCodes.TIME in keys:
            date_str = str(rows[YamlCodes.DATE])

            # Let pandas try the date separate first
            if len(date_str) == 6:
                date_str = pd.to_datetime(date_str).strftime('%Y-%m-%d')

            # Allow for nan time
            time_str = StringManager.parse_none(rows[YamlCodes.TIME])

            if time_str is not None:
                date_str += f" {time_str}"

            datetime = pd.to_datetime(date_str)

        elif YamlCodes.DATE in keys:
            datetime = pd.to_datetime(rows[YamlCodes.DATE])

        # Handle gpr data dates
        elif RowKeys.UTC_YEAR in keys and \
                RowKeys.UTC_DOY in keys and RowKeys.UTC_TOD in keys:
            base = pd.to_datetime(
                '{:d}-01-01 00:00:00 '.format(int(rows[RowKeys.UTC_YEAR])),
                utc=True
            )

            # Number of days since january 1
            days = int(rows[RowKeys.UTC_DOY]) - 1

            # Zulu time (time without colons)
            time = str(rows[RowKeys.UTC_TOD])
            hours = int(time[0:2])  # hours
            minutes = int(time[2:4])  # minutes
            seconds = int(time[4:6])  # seconds
            milliseconds = int(
                float('0.' + time.split('.')[-1]) * 1000
            )  # milliseconds

            delta = timedelta(
                days=days, hours=hours, minutes=minutes,
                seconds=seconds, milliseconds=milliseconds
            )
            # This is the only key set that ignores in_timezone
            datetime = base.astimezone(pytz.timezone('UTC')) + delta

        else:
            raise ValueError(f'Data is missing date/time info!\n{rows}')

        return datetime

    @staticmethod
    def adjust_timezone(date, in_timezone=None, out_timezone=None):
        """
        Adjust the datatime object for given timezone

        Args:
            date: parsed datetime object from pandas
            in_timezone: str - Incoming time zone
            out_timezone: str - Target time zone

        Returns:
            Date changed to target timezone
        """
        out_timezone = pytz.timezone(out_timezone)

        # Convert timezones if it is provided this variable gets rewritten
        # later
        if in_timezone is not None:
            in_tz = pytz.timezone(in_timezone)
        # Otherwise assume incoming data is the same timezone
        # TODO: how do we handle row based timezone
        else:
            raise ValueError("We did not receive a valid in_timezone")

        if date.tz is not None and date.tz.zone == in_tz.zone:
            date = date.astimezone(out_timezone)
        else:
            date = date.tz_localize(in_tz)
            date = date.astimezone(out_timezone)

        return date



from insitupy.campaigns.snowex import SnowExProfileData
from insitupy.profiles.base import ProfileData


class TestSnowExCampaign:
    def test_inheritance(self):
        assert issubclass(SnowExProfileData, ProfileData) is True



import logging
from typing import Tuple, Union

import utm

from .yaml_codes import YamlCodes

LOG = logging.getLogger(__name__)


class LocationManager:
    NORTHERN_HEMISPHERE = True

    @classmethod
    def parse_from_headers(cls, headers: dict) -> Tuple[
        float, float, Union[str, None], Union[str, None]
    ]:
        """
        Initial parse of lat, lon, easting, northing from the headers object

        Args:
            headers (dict): Parsed header with key-value pairs

        Returns:
            (Tuple) Latitude, Longitude, Easting, Northing
        """
        latitude = headers.get(YamlCodes.LATITUDE, None)
        longitude = headers.get(YamlCodes.LONGITUDE, None)
        easting = headers.get(YamlCodes.EASTING, None)
        northing = headers.get(YamlCodes.NORTHING, None)

        latitude = float(latitude) if latitude is not None else latitude
        longitude = float(longitude) if longitude is not None else longitude

        return latitude, longitude, easting, northing

    @classmethod
    def lat_lon_from_easting_northing(
        cls, headers: dict, easting: str, northing: str
    ) -> Tuple:
        """
        Convert Easting and Northing information to Lat/Lon objects via utm

        Args:
            headers (dict): Parsed header with key-value pairs
            easting (str): Detected Easting information
            northing (str): Detected Northing information

        Returns:
            (Tuple) Latitude, Longitude objects
        """
        zone_number = cls.parse_utm_epsg(headers)

        if isinstance(zone_number, str):
            raise RuntimeError(f"{zone_number} should be an integer")

        # Get the last two digits
        zone_number = int(str(zone_number)[-2:])
        try:
            lat, lon = utm.to_latlon(
                float(easting),
                float(northing),
                int(zone_number),
                northern=cls.NORTHERN_HEMISPHERE
            )
        except Exception as e:
            LOG.error(e)
            raise RuntimeError(f"Failed with {easting}, {northing}")

        return lat, lon

    @classmethod
    def parse(cls, headers: dict) -> Tuple[str, str, str, str]:
        """
        Parse the lat and lon from the header object

        Also parse the easting and northing
        # UTM Zone,13N
        # Easting,329131
        # Northing,4310328
        # Latitude,38.92524
        # Longitude,-106.97112
        # Flags,

        returns lat, lon, easting, northing
        """
        lat, lon, easting, northing = cls.parse_from_headers(headers)

        # Do nothing first
        if lat and lon:
            LOG.info("Latitude and Longitude parsed from the file header")
        elif easting and northing:
            LOG.info(
                "Latitude and Longitude converted from easting and northing"
            )
            lat, lon = cls.lat_lon_from_easting_northing(
                headers, easting, northing
            )
        else:
            raise ValueError(f"Could not parse location from {headers}")

        return lat, lon, easting, northing

    @classmethod
    def parse_utm_epsg(cls, headers: dict) -> Union[int, None]:
        # TODO: headers based utm?
        if YamlCodes.UTM_ZONE in headers.keys():
            utm_zone = int(
                ''.join(
                    [c for c in headers[YamlCodes.UTM_ZONE] if c.isnumeric()]
                )
            )
            return int(f"{YamlCodes.UTM_EPSG_PREFIX}{utm_zone}")
        elif YamlCodes.EPSG in headers.keys():
            return int(headers[YamlCodes.EPSG])
        else:
            LOG.warning(
                f"Could not find UTM keys in headers: {headers.keys()}"
            )
            return None



import logging
from pathlib import Path
from typing import List, Tuple, Optional, Union

import pandas as pd

from .dates import DateTimeManager
from .locations import LocationManager
from .strings import StringManager
from .yaml_codes import YamlCodes
from insitupy.variables import (
    ExtendableVariables,
    base_metadata_variables_yaml, base_primary_variables_yaml
)
from insitupy.profiles.metadata import ProfileMetaData

LOG = logging.getLogger(__name__)


class MetaDataParser:
    """
    Base class for parsing metadata in the header and data column names
    """
    DEFAULT_METADATA_VARIABLE_FILES = [base_metadata_variables_yaml]
    DEFAULT_PRIMARY_VARIABLE_FILES = [base_primary_variables_yaml]
    OUT_TIMEZONE = "UTC"
    DEFAULT_HEADER_SEPARATOR = ","
    DEFAULT_HEADER_LINE_START = '#'

    def __init__(
        self,
        timezone: Optional[str] = OUT_TIMEZONE,
        primary_variable_file: Optional[Union[str, Path]] = None,
        metadata_variable_file: Optional[Union[str, Path]] = None,
        header_sep=DEFAULT_HEADER_SEPARATOR,
        allow_split_lines: bool = False,
        allow_map_failures: bool = False,
        _id: Optional[str] = None,
        campaign_name: Optional[str] = None,
        units_map: Optional[dict] = None
    ):
        """
        Args:
            timezone: string timezone
            primary_variable_file:
                Path to file with primary variables mappings overwrites
            metadata_variable_file:
                Path to file with metadata variables mappings overwrites
            header_sep: expected header separator
            allow_split_lines: Allow for split header lines that
                don't start with the expected header character. In this case
                the number of header lines will be the max line starting with
                the expected character, and lines that don't start with
                that character will be combined with the previous line
            allow_map_failures: if a mapping fails, warn us and use the
                original string (default False)
            _id: optional pass in to override id in parse_id
            campaign_name: optional override for campaign name
            units_map = optional map of variable type to MeasurementDescription
        """
        self._allow_split_header_lines = allow_split_lines
        self._input_timezone = timezone
        self._header_sep = header_sep
        self._rough_obj = {}
        self._lat_lon_easting_northing = None
        self._id = _id
        self._campaign_name = campaign_name
        self._units_map = units_map or {}

        self.primary_variables = self.extend_variables(
            self.DEFAULT_PRIMARY_VARIABLE_FILES,
            primary_variable_file,
            allow_map_failures=allow_map_failures
        )
        self.metadata_variables = self.extend_variables(
            self.DEFAULT_METADATA_VARIABLE_FILES,
            metadata_variable_file,
            allow_map_failures=allow_map_failures,
        )

    @staticmethod
    def extend_variables(
        default: list,
        additions: Optional[Union[str, Path]] = None,
        allow_map_failures: bool = False
    ) -> ExtendableVariables:
        """
        Extends a list of default variables with optional additional entries
        and wraps them into an ExtendableVariables object. Identical code
        entries from the additions will overwrite the default entries.

        Args:
            default (list): A list of default variable entries
            additions (str | Path, optional):
                Additional entries from a file. (Default: None)
            allow_map_failures (bool, optional):
                Allow mapping failure in ExtendedVariables mapping.
                (Defaults: False)

        Returns:
            ExtendableVariables object mapping default and extended file
            entries.
        """
        entries = default + [additions] if additions else default
        return ExtendableVariables(
            entries=entries,
            allow_map_failures=allow_map_failures
        )

    @property
    def rough_obj(self):
        return self._rough_obj

    @property
    def units_map(self):
        return self._units_map

    @property
    def lat_lon_easting_northing(self):
        if self._lat_lon_easting_northing is None:
            self._lat_lon_easting_northing = \
                LocationManager.parse(self.rough_obj)
        return self._lat_lon_easting_northing

    # TODO: See remark on .parse_campaign_name
    def parse_id(self) -> str:
        if self._id is not None:
            return self._id
        else:
            try:
                return self.rough_obj[YamlCodes.ID_NAME]
            except KeyError:
                raise RuntimeError(f"Failed to parse ID from {self.rough_obj}")

    def parse_date_time(self) -> pd.Timestamp:
        datetime = DateTimeManager.parse(self.rough_obj)

        return DateTimeManager.adjust_timezone(
            datetime,
            in_timezone=self._input_timezone,
            out_timezone=self.OUT_TIMEZONE
        )

    def parse_latitude(self) -> float:
        return self.lat_lon_easting_northing[0]

    def parse_longitude(self) -> float:
        return self.lat_lon_easting_northing[1]

    def parse_utm_epsg(self) -> int:
        return LocationManager.parse_utm_epsg(self.rough_obj)

    # TODO: This needs to be revisited. Right now we have a mix and match
    #       of wording for a site vs a campaign
    def parse_campaign_name(self) -> str:
        if self._campaign_name is not None:
            return self._campaign_name
        else:
            try:
                return self.rough_obj[YamlCodes.SITE_NAME]
            except KeyError:
                raise RuntimeError(
                    f"Failed to parse Site Name from {self.rough_obj}"
                )

    # TODO: Expand base metadata YAML to detect this
    def parse_flags(self):
        result = None
        for k, v in self.rough_obj.items():
            if k in ["flags"]:
                result = v
                break

        return result

    # TODO: Expand base metadata YAML to detect this
    def observers_from_row(self, row):
        result = None
        for k, v in row.items():
            if k in ["observers"]:
                entry = v
                result = entry.split(", ")
                result = [r.strip() for r in result]
                break
        return result

    def parse_observers(self) -> List[str]:
        return self.observers_from_row(self.rough_obj)

    def _preparse_meta(self, meta_lines):
        """
        Organize the header lines into a dictionary with lower case keys
        """
        # Key value pairs are separate by some separator provided.
        data = {}

        # Collect key value pairs from the information above the column header
        for ln in meta_lines:
            d = ln.split(self._header_sep)

            # Key is always the first entry in comma sep list
            key = StringManager.standardize_key(d[0])

            # Avoid splitting on times
            if 'time' in key or 'date' in key:
                value = ':'.join(d[1:]).strip()
            else:
                value = ', '.join(d[1:])
                value = StringManager.clean_str(value)

            # cast the rough object key to a known key
            known_name, k_mapping = self.metadata_variables.from_mapping(key)

            # Assign non-empty strings to dictionary
            if key and value:
                data[known_name] = value.strip(
                    ' '
                ).replace('"', '').replace('  ', ' ')

            elif key and not value:
                data[known_name] = None
        return data

    def parse(self, filename: str) -> Tuple:
        """
        Parse the file and return a metadata object.
        We can override these methods as needed to parse the different
        metadata

        This populates self.rough_obj

        Args:
            filename: (str) Full path to the file with the header info to parse

        Returns:
            Tuple: metadata object, column list, position of header in file
        """
        meta_lines, columns, columns_map, header_position = \
            self.find_header_info(filename)
        self._rough_obj = self._preparse_meta(meta_lines)
        # Create a standard metadata object
        metadata = ProfileMetaData(
            site_name=self.parse_id(),
            date_time=self.parse_date_time(),
            latitude=self.parse_latitude(),
            longitude=self.parse_longitude(),
            utm_epsg=str(self.parse_utm_epsg()),
            campaign_name=self.parse_campaign_name(),
            flags=self.parse_flags(),
            observers=self.parse_observers()
        )

        return metadata, columns, columns_map, header_position

    def _parse_header(self, lines):
        # Key value pairs are separate by some separator provided.
        data = {}

        # Collect key value pairs from the information above the column header
        for ln in lines:
            d = ln.split(self._header_sep)

            # Key is always the first entry in comma sep list
            k = StringManager.standardize_key(d[0])

            # Avoid splitting on times
            if 'time' in k or 'date' in k:
                value = ':'.join(d[1:]).strip()
            else:
                value = ', '.join(d[1:])
                value = StringManager.clean_str(value)

            # Assign non empty strings to dictionary
            if k and value:
                data[k] = value.strip(' ').replace('"', '').replace('  ', ' ')

            elif k and not value:
                data[k] = None

        LOG.debug(
            'Discovered {} lines of valid header info.'
            ''.format(len(data.keys()))
        )
        return data

    def _parse_columns(self, str_line):
        """
        Parse the column names from the input line. This can include mapping
        """
        # Parse the columns header based on the size of the last line
        # Remove units
        # for c in ['()', '[]']:
        #     str_line = StringManager.strip_encapsulated(str_line, c)

        raw_cols = str_line.strip(
            self.DEFAULT_HEADER_LINE_START
        ).split(
            self._header_sep
        )
        # Clean the raw columns
        standard_cols = [StringManager.standardize_key(c) for c in raw_cols]
        # Infer units from the raw columns
        inferred_units = [
            StringManager.infer_unit_from_key(c) for c in raw_cols
        ]
        final_cols = []
        final_col_map = {}
        inferred_units_map = {}
        # Iterate through the columns and map to desired result
        for column, unit in zip(standard_cols, inferred_units):
            # TODO: could we return unmapped columns here?
            mapped_col, col_map = self.primary_variables.from_mapping(column)
            # Store the list of columns to use when reading in the
            # dataframe
            final_cols.append(mapped_col)
            # Store the map of a column name to the known variable
            final_col_map = {**final_col_map, **col_map}
            # Store the map of column name to inferred unit
            result_obj = col_map[mapped_col]
            if result_obj is None:
                if self.primary_variables.allow_map_failures:
                    LOG.warning(f"No unit for {column}")
                else:
                    raise RuntimeError(
                        f"No unit for {column} - column mapping has failed"
                    )
            else:
                # override the code if it is given
                inferred_units_map[
                    result_obj.code
                ] = self._units_map.get(result_obj.code) or unit

        return final_cols, final_col_map, inferred_units_map

    def find_header_info(self, filename: str):
        """
        Read in all site details file for a pit If the filename has the word
        site in it then we read everything in the file. Otherwise, we use this
        to read all the site data up to the header of the profile.

        E.g. Read all commented data until we see a column descriptor.

        Args:
            filename: Path to a csv containing # leading lines with site details

        Returns:
            tuple: **data** - Dictionary containing site details
                   **columns** - List of clean column names
                   **header_pos** - Index of the columns header for skiprows in
                                    read_csv
       """
        filename = str(filename)
        try:
            with open(filename, "r", encoding="utf-8-sig") as f:
                lines = f.readlines()
        except UnicodeDecodeError:
            with open(filename, "r", encoding="latin1") as f:
                lines = f.readlines()

        # Site description files have no need for column lists
        if 'site' in filename.lower():
            LOG.info('Parsing site description header...')
            columns = None
            header_pos = None
            header_indicator = None
            columns_map = {}
            self._units_map = self._units_map

        # Find the column names and where it is in the file
        else:
            header_pos, header_indicator = self._find_header_position(lines)
            # identify columns, map columns, and units map
            columns, columns_map, units_map = self._parse_columns(
                lines[header_pos]
            )
            # Combine with user defined units map
            self._units_map = {**self._units_map, **units_map}
            LOG.debug(
                f'Column Data found to be {len(columns)} columns based on'
                f' Line {header_pos}'
            )
            # Only parse what we know if the header
            lines = lines[0:header_pos]

        final_lines = lines

        # Clean up the lines from line returns to grab header info
        final_lines = [ln.strip() for ln in final_lines]
        # Join all data and split on header separator
        # This handles combining split lines
        str_data = " ".join(final_lines).split(self.DEFAULT_HEADER_LINE_START)
        str_data = [ln.strip() for ln in str_data if ln]

        return str_data, columns, columns_map, header_pos

    def _iterative_header_pos_search(self, lines, n_columns, header_indicator):
        # Use these to monitor if a larger column count is found
        header_pos = 0
        for i, l in enumerate(lines):
            if i == 0:
                previous = StringManager.get_alpha_ratio(lines[i])
            else:
                previous = StringManager.get_alpha_ratio(lines[i - 1])

            if StringManager.line_is_header(
                l, expected_columns=n_columns,
                header_indicator=header_indicator,
                previous_alpha_ratio=previous
            ):
                header_pos = i

            if i > header_pos:
                break
        return header_pos

    def _find_header_position(self, lines):
        """
        A flexible method that attempts to find and standardize column names
        for csv data. Looks for a comma separated line with N entries == to the
        last line in the file. If an entry is found with more commas than the
        last line then we use that. This allows us to have data that doesn't
        have all the commas in the data (SSA typically missing the comma for
        veg unless it was notable)

        Assumptions:

        1. The last line in file is of representative csv data
        2. The header is the last column that has more chars than numbers

        Args:
            lines: Complete list of strings from the file

        Returns:
            header position
        """

        # Minimum column size should match the last line of data (Assumption
        # #2)
        n_columns = len(lines[-1].split(','))

        if lines[0][0] == self.DEFAULT_HEADER_LINE_START:
            header_indicator = self.DEFAULT_HEADER_LINE_START
        else:
            header_indicator = None

        if self._allow_split_header_lines:
            if header_indicator is None:
                raise RuntimeError(
                    "Cannot allow split lines with no clear header indicator"
                )
            else:
                # header pos is max lines with
                # first character == header indicator
                header_indices = [
                    index for index, value in enumerate(lines)
                    if value[0] == header_indicator
                ]
                header_pos = max(header_indices)

        else:
            header_pos = self._iterative_header_pos_search(
                lines, n_columns, header_indicator
            )

        LOG.debug('Found end of header at line {}...'.format(header_pos))
        return header_pos, header_indicator



class YamlCodes:
    """
    These are the standardized codes mapped in the YAML files that
    detect variants.
    """
    DATE = 'date'
    DATE_TIME = "datetime"
    EASTING = "easting"
    EPSG = 'epsg'
    ID_NAME = "pit_id"
    LATITUDE = "latitude"
    LONGITUDE = "longitude"
    NORTHING = "northing"
    SITE_NAME = 'site_name'
    TIME = 'time'
    UTM_EPSG_PREFIX = "269"
    UTM_ZONE = 'utm_zone'
